import json
import typing

import chromadb
import networkx as nx
from chromadb import API
from chromadb.api.models.Collection import Collection
from chromadb.utils.embedding_functions import SentenceTransformerEmbeddingFunction
from loguru import logger
from networkx import Graph
from pydantic import BaseModel
from pydantic_settings import BaseSettings
from tqdm import tqdm

from srctag.model import FileContext, RuntimeContext, SrcTagException


class StorageDoc(BaseModel):
    document: str
    metadata: typing.Dict[str, str]
    id: str


class MetadataConstant(object):
    KEY_SOURCE = "source"
    KEY_COMMIT_SHA = "commit_sha"
    KEY_DATA_TYPE = "data_type"
    KEY_ISSUE_ID = "issue_id"
    KEY_TAG = "tag"

    # use in chroma
    DATA_TYPE_COMMIT_MSG = "commit_msg"
    DATA_TYPE_ISSUE = "issue"


class StorageConfig(BaseSettings):
    db_path: str = ""
    collection_name: str = "default_collection"

    # English: paraphrase-MiniLM-L6-v2
    # Multi langs: paraphrase-multilingual-MiniLM-L12-v2
    st_model_name: str = "paraphrase-MiniLM-L6-v2"

    # content mapping for avoiding too much I/O
    # "#11" -> "content for #11"
    issue_mapping: typing.Dict[str, str] = dict()

    data_types: typing.Set[str] = {MetadataConstant.DATA_TYPE_COMMIT_MSG, MetadataConstant.DATA_TYPE_ISSUE}

    def load_issue_mapping_from_gh_json_file(self, gh_json_file: str):
        with open(gh_json_file) as f:
            content = json.load(f)
        assert isinstance(content, list), "not a valid issue dump"

        for each in content:
            sharp_id = f'#{each["number"]}'
            self.issue_mapping[sharp_id] = each["title"]
        logger.info(f"load {len(content)} issues from {gh_json_file}")


class Storage(object):
    def __init__(self, config: StorageConfig = None):
        if not config:
            config = StorageConfig()
        self.config = config

        self.chromadb: typing.Optional[API] = None
        self.chromadb_collection: typing.Optional[Collection] = None
        self.relations: Graph = nx.Graph()

    def init_chroma(self):
        if self.chromadb and self.chromadb_collection:
            return

        if self.config.db_path:
            self.chromadb = chromadb.PersistentClient(path=self.config.db_path)
        else:
            # by default, using in-memory db
            self.chromadb = chromadb.Client()

        self.chromadb_collection = self.chromadb.get_or_create_collection(
            self.config.collection_name,
            embedding_function=SentenceTransformerEmbeddingFunction(
                model_name=self.config.st_model_name
            ),
            # dis range: [0, 1]
            metadata={"hnsw:space": "l2"}
        )

    def process_commit_msg(self, file: FileContext, collection: Collection, _: RuntimeContext):
        """ can be overwritten for custom processing """
        targets = []
        for each in file.commits:
            # keep enough data in metadata for calc the final score
            item = StorageDoc(
                document=each.message,
                metadata={
                    MetadataConstant.KEY_SOURCE: file.name,
                    MetadataConstant.KEY_COMMIT_SHA: str(each.hexsha),
                    MetadataConstant.KEY_DATA_TYPE: MetadataConstant.DATA_TYPE_COMMIT_MSG,
                },
                id=f"{MetadataConstant.DATA_TYPE_COMMIT_MSG}|{file.name}|{each.hexsha}"
            )
            targets.append(item)

        for each in targets:
            collection.add(
                documents=[each.document],
                metadatas=[each.metadata],
                ids=[each.id],
            )

    def process_issue_id_to_title(self, issue_id: str) -> str:
        # easily reach the API limit if using server API here,
        # so we use issue_mapping, keep it simple
        return self.config.issue_mapping.get(issue_id, "")

    def process_issue(self, _: FileContext, collection: Collection, ctx: RuntimeContext):
        issue_id_list = [x for x, y in ctx.relations.nodes(data=True) if
                         y["node_type"] == MetadataConstant.KEY_ISSUE_ID]

        targets = []
        for each_issue_id in issue_id_list:
            each_issue_content = self.process_issue_id_to_title(each_issue_id)
            if not each_issue_content:
                continue

            item = StorageDoc(
                document=each_issue_content,
                metadata={
                    MetadataConstant.KEY_ISSUE_ID: each_issue_id,
                    MetadataConstant.KEY_DATA_TYPE: MetadataConstant.DATA_TYPE_ISSUE,
                },
                id=f"{MetadataConstant.DATA_TYPE_ISSUE}|{each_issue_id}"
            )
            targets.append(item)

            # END issue loop
        # END commit loop

        for each in targets:
            collection.upsert(
                documents=[each.document],
                metadatas=[each.metadata],
                ids=[each.id],
            )

    def process_file_ctx(self, file: FileContext, collection: Collection, ctx: RuntimeContext):
        process_dict = {
            MetadataConstant.DATA_TYPE_ISSUE: self.process_issue,
            MetadataConstant.DATA_TYPE_COMMIT_MSG: self.process_commit_msg
        }
        for each in self.config.data_types:
            if each not in process_dict:
                raise SrcTagException(f"invalid data type: {each}")
            process_dict[each](file, collection, ctx)

    def embed_file(self, file: FileContext, ctx: RuntimeContext):
        if not file.commits:
            logger.warning(f"no related commits found: {file.name}")
            return

        self.init_chroma()
        self.process_file_ctx(file, self.chromadb_collection, ctx)

    def embed_ctx(self, ctx: RuntimeContext):
        self.init_chroma()
        self.relations = ctx.relations
        logger.info("start embedding source files")
        for each_file in tqdm(ctx.files.values()):
            self.embed_file(each_file, ctx)
        logger.info("embedding finished")



import subprocess
import typing

import chromadb
import click
import networkx
import networkx as nx
from chromadb.utils.embedding_functions import SentenceTransformerEmbeddingFunction
from loguru import logger

from srctag.collector import Collector, FileLevelEnum
from srctag.storage import Storage, MetadataConstant
from srctag.tagger import Tagger


@click.group()
def cli():
    pass


@cli.command()
def prepare():
    """ usually used for pre-downloading sentence-transformer models """

    logger.info("Start checking env. It may takes a few minutes for downloading models ...")
    # try to embed
    chromadb_cli = chromadb.Client()
    collection = chromadb_cli.get_or_create_collection(
        "testonly",
        embedding_function=SentenceTransformerEmbeddingFunction(),
    )
    collection.add(
        documents=["doc"],
        ids=["id"],
    )
    assert collection.count() == 1
    click.echo("ok.")


@cli.command()
@click.option("--repo-root", default=".", help="Repository root directory")
@click.option("--max-depth-limit", default=-1, help="Maximum depth limit")
@click.option("--include-regex", default="", help="File include regex pattern")
@click.option("--tags-file", type=click.File("r"), default="./srctag.txt", help="Path to a text file containing tags")
@click.option("--output-path", default="", help="Output file path for CSV")
@click.option("--file-level", default=FileLevelEnum.FILE.value, help="Scan file level, FILE or DIR, default to FILE")
@click.option("--st-model", default="", help="Sentence Transformer Model")
@click.option("--commit-include-regex", default="", help="Commit message include regex pattern")
def tag(repo_root, max_depth_limit, include_regex, tags_file, output_path, file_level, st_model, commit_include_regex):
    """ tag your repo """
    collector = Collector()
    collector.config.repo_root = repo_root
    collector.config.max_depth_limit = max_depth_limit
    collector.config.include_regex = include_regex
    collector.config.file_level = file_level
    collector.config.commit_include_regex = commit_include_regex

    ctx = collector.collect_metadata()
    storage = Storage()
    if st_model:
        storage.config.st_model_name = st_model
    storage.embed_ctx(ctx)
    tagger = Tagger()

    assert tags_file, "no tag file provided"
    tags = [each.strip() for each in tags_file.read().splitlines()]

    tagger.config.tags = tags
    tag_dict = tagger.tag(storage)

    if output_path:
        tag_dict.export_csv(path=output_path)
    else:
        tag_dict.export_csv()


@cli.command()
@click.option("--repo-root", default=".", help="Repository root directory")
@click.option("--max-depth-limit", default=-1, help="Maximum depth limit")
@click.option("--include-regex", default="", help="File include regex pattern")
@click.option("--file-level", default=FileLevelEnum.FILE.value, help="Scan file level, FILE or DIR, default to FILE")
@click.option("--output-path", default="srctag.dot", help="Output file path for DOT")
@click.option("--issue-regex", default="", help="Issue regex")
def graph(repo_root, max_depth_limit, include_regex, file_level, output_path, issue_regex):
    """ create relations graph from your repo """
    collector = Collector()
    collector.config.repo_root = repo_root
    collector.config.max_depth_limit = max_depth_limit
    collector.config.include_regex = include_regex
    collector.config.file_level = file_level
    if issue_regex:
        collector.config.issue_regex = issue_regex

    ctx = collector.collect_metadata()
    relation_graph = ctx.relations
    render_dot(relation_graph, output_path)


@cli.command()
@click.option("--repo-root", default=".", help="Repository root directory")
@click.option("--max-depth-limit", default=-1, help="Maximum depth limit")
@click.option("--diff-target", default="HEAD~1", help="diff target rev")
@click.option("--file-level", default=FileLevelEnum.FILE.value, help="Scan file level, FILE or DIR, default to FILE")
@click.option("--output-path", default="srctag.dot", help="Output file path for DOT")
@click.option("--batch", default=1, help="")
@click.option("--issue-regex", default="", help="Issue regex")
def diff(repo_root, max_depth_limit, diff_target, file_level, output_path, batch, issue_regex):
    """ create relations graph from your repo, with diff """
    base_file_set = get_git_diff_files(diff_target)
    total_file_set = base_file_set
    logger.info(f"base file set: {base_file_set}")

    for i in range(batch):
        collector = Collector()
        collector.config.repo_root = repo_root
        collector.config.max_depth_limit = max_depth_limit
        collector.config.file_level = file_level
        collector.config.include_file_list = base_file_set
        if issue_regex:
            collector.config.issue_regex = issue_regex

        ctx = collector.collect_metadata()

        # enlarge this file set
        file_set = set(ctx.files.keys())
        # network graph
        for each in file_set:
            if not ctx.relations.has_node(each):
                logger.warning(f"node {each} not in graph")
                continue

            issues = set()
            for each_node in ctx.relations.neighbors(each):
                if ctx.relations.nodes[each_node]["node_type"] != MetadataConstant.KEY_ISSUE_ID:
                    continue
                issues.add(each_node)
            # END issue query

            for each_issue in issues:
                related_files = ctx.relations.neighbors(each_issue)
                file_set = file_set.union(related_files)

        # END file query

        logger.info(f"batch {i} end, files: {len(base_file_set)} -> {len(file_set)}")
        new_file_set = file_set - total_file_set
        if not new_file_set:
            logger.info(f"file range search ready: {len(total_file_set)}")
            break
        # update base scope, and run again
        total_file_set = file_set
        base_file_set = new_file_set
    # END loop batch

    collector = Collector()
    collector.config.repo_root = repo_root
    collector.config.max_depth_limit = max_depth_limit
    collector.config.file_level = file_level
    collector.config.include_file_list = total_file_set
    if issue_regex:
        collector.config.issue_regex = issue_regex
    ctx = collector.collect_metadata()
    relation_graph = ctx.relations
    render_dot(relation_graph, output_path)


def get_git_diff_files(target: str) -> typing.Set[str]:
    result = subprocess.check_output(['git', 'diff', target, '--name-only'], text=True)
    diff_files = result.splitlines()
    return set(diff_files)


def render_dot(relation_graph: networkx.Graph, output: str):
    node_colors = {
        MetadataConstant.KEY_SOURCE: 'tomato',
        MetadataConstant.KEY_ISSUE_ID: 'lightgreen',
        MetadataConstant.KEY_COMMIT_SHA: 'lightblue',
    }
    node_color_mapping = {node: node_colors.get(data.get('node_type', 'default'), 'lightgray') for node, data in
                          relation_graph.nodes(data=True)}
    for node, color in node_color_mapping.items():
        relation_graph.nodes[node]['color'] = color

    nx.drawing.nx_pydot.write_dot(relation_graph, output)
    logger.info("rendered graph to %s", output)


if __name__ == '__main__':
    cli()
